package org.qatools.validator;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import org.qatools.info.Info;

public abstract class AValidator {
	private static Info info = new Info();	// Automatically print blog info... 
	// unless it's deactivated prior to loading this class with QATools.disableIn()

	// -----------------------------------------------------------
	// ----- abstract methods
	// -----------------------------------------------------------
	public abstract void assertTrue(String message, boolean condition);
	public abstract void assertFalse(String message, boolean condition);
	public abstract void fail(String message);
	public abstract void assertEquals(String message, Object expected, Object actual);
	public abstract void assertEquals(String message, String expected, String actual);
	public abstract void assertEquals(String message, double expected, double actual, double delta);
	public abstract void assertEquals(String message, float expected, float actual, float delta);
	public abstract void assertEquals(String message, long expected, long actual);
	public abstract void assertEquals(String message, boolean expected, boolean actual);
  	public abstract void assertEquals(String message, byte expected, byte actual);
  	public abstract void assertEquals(String message, char expected, char actual);
	public abstract void assertEquals(String message, short expected, short actual);
  	public abstract void assertEquals(String message, int expected, int actual);
	public abstract void assertNotNull(String message, Object object);
	public abstract void assertNull(String message, Object object);
    public abstract void assertNull(String message, String object);
	public abstract <T extends Object> void assertListNotNullNotEmpty(String message, List<T> list);
	public abstract void assertStrNotNullNotEmpty(String message, String str);
	public abstract void assertGreaterThan(String message, long smallExpected, long bigActual);
	public abstract void assertGreaterThan(String message, int smallExpected, int bigActual);
	public abstract void assertGreaterThanEquals(String message, long smallExpected, long bigActual);
	public abstract void assertGreaterThanEquals(String message, int smallExpected, int bigActual);
    public abstract void assertLessThan(String message, long bigExpected, long smallActual);
    public abstract void assertLessThan(String message, int bigExpected, int smallActual);
    public abstract void assertLessThanEquals(String message, long bigExpected, long smallActual);
    public abstract void assertLessThanEquals(String message, int bigExpected, int smallActual);
	public abstract <T extends Object> void assertContains(String message, T expected, List<T> actual);
	public abstract <T extends Object> void assertContains(String message, List<T> expected, T actual);
	public abstract void assertStrContainsStr(String message, String longExpected, String shortActual);
	public abstract void assertStrContainsStr2(String message, String shortExpected, String longActual);
	public abstract void assertEquals(String message, List<Object> expected, List<Object> actual);
	public abstract void assertNotEquals(String message, Object expected, Object actual);
	public abstract void assertNotEquals(String message, String expected, String actual);
	public abstract void assertNotEquals(String message, long expected, long actual);
	public abstract void assertNotEquals(String message, boolean expected, boolean actual);
  	public abstract void assertNotEquals(String message, int expected, int actual);
	public abstract void assertRange(String message, int expectedLow, int expectedHigh, int actual);
	public abstract void assertRange(String message, long expectedLow, long expectedHigh, long actual);

    // ------------------------------------------------------
    // ----- static methods
    // ------------------------------------------------------

    /**
     * 
     * @param msg
     */
    public static void throwNewException(String msg) throws Exception {
        AValidator v = new BasicValidator();
        v.addError(msg);
        v.throwExceptionIfErrors();
    }


    /**
     * 
     * @param msg
     */
    public static void throwNewException(Exception ex) throws Exception {
        String msg = null;
        if (ex == null) {
        	msg = "A NULL Exception was received";
        } else if (ex.getLocalizedMessage() != null) {
        	msg = ex.getLocalizedMessage();
        } else if (ex.getMessage() != null) {
        	msg = ex.getMessage();
        } else {
        	msg = "throwNewException";
        }
        AValidator v = new BasicValidator();
        v.addError(msg);
        v.throwExceptionIfErrors();
    }
    

    /**
     * 
     * @param msg
     */
    public static void throwNewRuntimeException(String msg) {
        AValidator v = new BasicValidator();
        v.addError(msg);
        v.throwRuntimeExceptionIfErrors();
    }
    
    

    /**
     * 
     * @param msg
     */
    public static void throwNewRuntimeException(Exception ex) {
        String msg = null;
        if (ex == null) {
        	msg = "A NULL Exception was received";
        } else if (ex.getLocalizedMessage() != null) {
        	msg = ex.getLocalizedMessage();
        } else if (ex.getMessage() != null) {
        	msg = ex.getMessage();
        } else {
        	msg = "throwNewRuntimeException";
        }
        AValidator v = new BasicValidator();
        v.addError(msg);
        v.throwRuntimeExceptionIfErrors();
    }
    
    
    
    
    // ------------------------------------------------------
    // ----- static initialization
    // ------------------------------------------------------

    static {
    	//addPrintFormat(PrintFormat.BR);				// error messages contain <br> for html reports
    	addPrintFormat(PrintFormat.NEWLINE);		// error messages contain \n
        addPrintFormat(PrintFormat.NUMBER_ERRORS);  // errors are numbered
        //addPrintFormat(PrintFormat.HASH_MARKS);  	// add headers to separate errors & debugs with hash marks
    	
    	
        //
        // control how much error & debug info shows up in the surefire report
        // allows you to decrease how much info shows up in your build systems
    	setErrorOption(ErrorOption.SUREFIRE_FULL);	// ErrorOption enum allows you to control how much info shows up in the surefire report
    	setDebugOption(DebugOption.EVERYWHERE);		// DebugOption enum allows you to control how much debug info shows up in the surefire report
    }

    // ------------------------------------------------------
    // ----- class variables
    // ------------------------------------------------------

    //errors
    protected List<String> errors = null;
    
    //debug
    protected List<String> debugs = null;
    
    // print formats
	private static List<PrintFormat> sPrintFormats;

    // error options - formatting errors
	protected static ErrorOption sErrorOption;
	
    // debug options - formatting debug messages
	protected static DebugOption sDebugOption;
	
	//
	// String Processing Options
	private static List<StringCleaningOption> sStringCleaningOptions;	// trim, remove dup spaces, null equivalence
	private static List<String> sNullEquivalenceStrings;					// customize null = "", " ", "null", etc - must be used with sStringProcessingOptions.NULL_EQUIVALENCE
	private static boolean sStringCleaningForReportFlag = true;	//so reports aren't cluttered by really long strings
		//TODO - could change this to concatLongStringsForReport(int length)

    // ------------------------------------------------------
    // ----- reset
    // ------------------------------------------------------

    /**
     * Resets there error & debug lists
     */
    public void reset() {
        errors = null;
        debugs = null;
    }

    // ------------------------------------------------------
    // ----- Print Format
    // ------------------------------------------------------


    /**
     * Allows you to set the print format suitable for your environment
     *
     *     	PrintFormat.NEWLINE - errors show up on a separate line
     *     
     *     	PrintFormat.BR - html formatted <br> is added to each error for viewing in a browser
     *     
     *    	PrintFormat.NUMBER_ERRORS - number errors
     *    
     *    	PrintFormat.HASH_MARKS - section headers for errors and debugs
     * 
     * @param printFormats
     */
    public static void setPrintFormats(List<PrintFormat> printFormats) {
    	if (printFormats == null || printFormats.size() == 0) {
    		sPrintFormats = null;
    		return;
    	}
    	
    	//
    	// The reason we don't just do this assignment sPrintFormats = printFormats
    	// is Arrays.asList(....)  creates an immutable list
    	sPrintFormats = null;	// over -rides the default settings
    	for (PrintFormat pf : printFormats) {
    		addPrintFormat(pf);
    	}
	}

    /**
     * Allows you to set the print format suitable for your environment
     *
     *     	PrintFormat.NEWLINE - errors show up on a separate line
     *     
     *     	PrintFormat.BR - html formatted <br> is added to each error for viewing in a browser
     *     
     *    	PrintFormat.NUMBER_ERRORS - number errors
     *    
     *    	PrintFormat.HASH_MARKS - section headers for errors and debugs
     * 
     * @param printFormats
     */
    public static void addPrintFormat(PrintFormat printFormat) {
    	if (sPrintFormats == null || sPrintFormats.size()==0) {
    		sPrintFormats = new LinkedList<PrintFormat>();
    	}
    	sPrintFormats.add(printFormat);
	}
    
    /**
     *     	ErrorOption.SUREFIRE_FULL - configure report to show all details in your surefire reports
     *     
     *     	ErrorOption.SUREFIRE_SHORT - only show the first error in your surefire reports. 

     * 
     * @param printFormat
     */
    public static void setErrorOption(ErrorOption errorOption) {
    	sErrorOption = errorOption;
	}

    /**
     *     DebugOption.EVERYWHERE - debug details show up in surefire reports and QATools reports
     *    
     *     DebugOption.QATOOLS_DETAIL_REPORT_ONLY - only show debug details in QATools reports. 
     *     Debug messages do not show up in surefire reports.
     * 
     * 
     * @param printFormat
     */
    public static void setDebugOption(DebugOption debugOption) {
    	sDebugOption = debugOption;
	}

    
    /**
     * @return
     */
    private boolean isNewline() {
    	if (sPrintFormats.contains(PrintFormat.NEWLINE)) {
    		return true;
    	}
    	return false;
    }

    /**
     * @return
     */
    private boolean isBR() {
    	if (sPrintFormats.contains(PrintFormat.BR)) {
    		return true;
    	}
    	return false;
    }

    /**
     * @return
     */
    private boolean isNumberErrors() {
        if (sPrintFormats.contains(PrintFormat.NUMBER_ERRORS)) {
            return true;
        }
        return false;
    }

    /**
     * @return
     */
    private boolean isHash() {
        if (sPrintFormats.contains(PrintFormat.HASH_MARKS)) {
            return true;
        }
        return false;
    }

    /**
     * 
     */
    protected void formatErrors() {
    	if (!doErrorsExist()) {
    		return;
    	}
    	String br = "";
    	String newline = "";
    	for (int i=0; i<errors.size(); i++) {
    		String str = errors.get(i);
    		int b = i + 1;
    		
    		if (isNumberErrors()) {
    			str = "(" + b + ") " + str;
    		}	//isNumberErrors

    		if (isBR()) {
    			br = "<br>";
    			str = "<br>" + str;
    		}	//isBR

    		if (isNewline()) {
    			newline = "\n";
    			str = "\n" + str;
    		}	//isNewline

    		errors.set(i, str);
    	}	//for
    	
    	//
    	//only need this extra header info if there is also debug info
    	if (doDebugsExist()) {
    	    String hashMarks = (isHash() ? newline + br + "##################################" : "");
        	errors.add(0, hashMarks);
        	errors.add(0, newline + br + "### Error details");
            errors.add(0, hashMarks);
    	}

    }	//method

    /**
     * 
     */
    protected void formatDebugs() {    	
    	if (!doDebugsExist()) {
    		return;
    	}
    	String br="";
    	String newline = "";
    	for (int i=0; i<debugs.size(); i++) {
    		String str = debugs.get(i);
    		int b = i + 1;
    		
    		if (isNumberErrors()) {
    			str = "(" + b + ") " + str;
    		}	//isNumberErrors

    		if (isBR()) {
    			str = "<br>" + str;
    			br = "<br>";
    		}	//isBR

    		if (isNewline()) {
    			str = "\n" + str;
    			newline = "\n";
    		}	//isNewline

    		debugs.set(i, str);
    	}	//for
        String hashMarks = (isHash() ? newline + br + "##################################" : "");
        debugs.add(0, hashMarks);
		debugs.add(0, newline + br + "### Debug / Contextual Information");
		debugs.add(0, hashMarks);

    }	//method

    
	// ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------

	
	String formatPassMsg(String message, Object expected, Object actual) {
		String formatted= "";
		if (message != null)
			formatted= message+" ";
		return formatted+"expected:'"+expected+"' and was:'"+actual+"'";
	}
	String formatErrorMsg(String message, Object expected, Object actual) {
		String formatted= "";
		if (message != null)
			formatted= message+" ";
		return formatted+"expected:'"+expected+"' but was:'"+actual+"'";
	}


    // ------------------------------------------------------
    // ----- StringProcessingOption
    // ------------------------------------------------------


	
	/**
	 * 		StringProcessingOption.NULL_EQUIVALENCE enables null equivalents processing. See setNullEquivalenceString
	 * 
	 * 		StringProcessingOption.REMOVE_DUPLICATE_SPACES removes duplicate spaces from strings
	 * 
	 * 		StringProcessingOption.TRIM removies leading and trailing spaces from string
	 * 
	 * @param stringProcessingOption
	 */
	public static void setStringCleaningOption(StringCleaningOption stringProcessingOption) {
	    //System.out.println("\nEnter setStringProcessingOption " + stringProcessingOption + "\n");
		
		if (stringProcessingOption == null) {
		    //System.out.println ("stringProcessingOption == null");
			setStringCleaningOptions(null);
		}
		setStringCleaningOptions(Arrays.asList(stringProcessingOption));
	}
	
	/**
	 * 		StringCleaningOption.NULL_EQUIVALENCE enables null equivalents processing. See setNullEquivalenceString
	 * 
	 * 		StringCleaningOption.REMOVE_DUPLICATE_SPACES removes duplicate spaces from strings
	 * 
	 * 		StringCleaningOption.TRIM removies leading and trailing spaces from string
	 * 
	 * @param stringCleaningOptions
	 */
	public static void setStringCleaningOptions(List<StringCleaningOption> stringCleaningOptions) {
	    if (stringCleaningOptions==null || 0 == stringCleaningOptions.size ()) {
	        sStringCleaningOptions = null;
	        //System.out.println("sStringProcessingOptions = null;");
	    } else {
            sStringCleaningOptions = new LinkedList<StringCleaningOption>();
            for (StringCleaningOption option : stringCleaningOptions) {
                sStringCleaningOptions.add (option);
                //System.out.println("adding " + option);
            }
	    }

	}
	
	
	/**
	 * 		StringCleaningOption.NULL_EQUIVALENCE enables null equivalents processing. See setNullEquivalenceString
	 * 
	 * 		StringCleaningOption.REMOVE_DUPLICATE_SPACES removes duplicate spaces from strings
	 * 
	 * 		StringCleaningOption.TRIM removies leading and trailing spaces from string
	 * 
	 * @param stringProcessingOptions
	 */
    public static void addStringCleaningOption(StringCleaningOption stringCleaningOption) {
        if (sStringCleaningOptions == null) {
            sStringCleaningOptions = new LinkedList<StringCleaningOption>();
        }
        sStringCleaningOptions.add (stringCleaningOption);
    }

	
	/**
	 * Sets values to be used for StringProcessingOption.NULL_EQUIVALENCE. 
	 * Automatically enables StringProcessingOption.NULL_EQUIVALENCE if not already enabled
	 *  
	 * For example, "", " ", "{}", "null", "0.0" can be set equivalent to null so that the Validator treats your values as equivalent to null
	 */
	public static void setNullEquivalenceString(String str) {
		if (str == null) {
			setNullEquivalenceStrings(null);
		}
		setNullEquivalenceStrings(Arrays.asList(str));
	}

	/**
	 * Sets values to be used for StringProcessingOption.NULL_EQUIVALENCE. 
	 * Automatically enables StringProcessingOption.NULL_EQUIVALENCE if not already enabled
	 *  
	 * For example, "", " ", "{}", "null", "0.0" can be set equivalent to null so that the Validator treats your values as equivalent to null
	 */
	public static void setNullEquivalenceStrings(List<String> strings) {
	    if (strings==null || 0 == strings.size ()) {
	        sNullEquivalenceStrings = null;
	    } else {
            addStringCleaningOption(StringCleaningOption.NULL_EQUIVALENCE);   // this is also required to do null equiv processing          
	        sNullEquivalenceStrings = new LinkedList<String>();
	        for (String s : strings) {
	            //System.out.println("setNullEquivalenceStrings add - " + s);
	        	if (s != null) {
		            sNullEquivalenceStrings.add (s);
	        	}
	        }
	    }
	}

    // ------------------------------------------------------
    // ----- methods for processing (cleaning up) strings
    // ------------------------------------------------------

	/**
	 * Returns Object[]
	 * 		Boolean.TRUE / FALSE if the string was changed
	 * 		original value
	 * 		new clean value
	 * @param str
	 * @return
	 */
	private static Object[] cleanStringBase(String str) {
		Object[] result = new Object[3];
		result[0] = Boolean.FALSE;	// assume string does not need cleaning
		result[1] = str;			// this element contains the original value
		result[2] = null;			//doesn't matter unless processing occurs
		
		if (str == null) {
	        //System.out.println("processString returning input = " + str);
			return result;
		}

        //System.out.println("processString input = '" + str + "'");

        if (sStringCleaningOptions==null) {
	        //System.out.println("sStringProcessingOptions==null");
			return result;
		} 
        
        
		//
        // trim string
		String s = str;
		if (sStringCleaningOptions.contains(StringCleaningOption.TRIM)) {
            //System.out.println("processString in trim(1) s = '" + s + "'");
			s = s.trim();
            //System.out.println("processString in trim(2) s = '" + s + "'");
		}
		
		
		//
		// remove double spaces
		if (sStringCleaningOptions.contains(StringCleaningOption.REMOVE_DUPLICATE_SPACES)) {
		    String SINGLE_SPACE = " ";
		    String DOUBLE_SAPCE = "  ";
			while(s.contains(DOUBLE_SAPCE)) {
				s = s.replaceAll(DOUBLE_SAPCE, SINGLE_SPACE);
			}
		}
		
		
		//
		// Null equivalence changes these
		//		"" to null
		//		" " to null
		//		"0.0" to null
		if (sStringCleaningOptions.contains(StringCleaningOption.NULL_EQUIVALENCE)) {
		    //System.out.println("Yes - NULL_EQUIVALENCE");
		    if (sNullEquivalenceStrings!=null) {
		        //System.out.println("Yes - sNullEquivalenceStrings!=null");
		        if (0<sNullEquivalenceStrings.size()) {
		            //System.out.println("Yes - 0<sNullEquivalenceStrings.size()");
		            for (String b : sNullEquivalenceStrings) {
		                //System.out.println("null equiv setting = " + b + ", s = " + s);
		                if (b!=null && b.equals(s)) {
		                    s = null;
		                    //System.out.println("set s = null");
		                } // 
		            } //for
		        } //if 0<sNullEquivalenceStrings.size()
		    } // if sNullEquivalenceStrings!=null
		} // if NULL_EQUIVALENCE
		
		if (!str.equals(s)) {
			result[0] = Boolean.TRUE;
			result[2] = s;
		}
		
		return result;		
	}

	/**
	 * 
	 * @param str
	 * @return
	 */
	protected static String cleanString(String str) {
		Object[] result = cleanStringBase(str);
		if (Boolean.TRUE.equals(result[0])) {
			return (String)result[2];	//the modified value
		}
		return (String) result[1];	//the original value
	}

	/**
	 * Convenience method for better reporting. The detail report will show the original & new value of any strings that are processed.
	 * 
	 * If processing modified the original string, this returns
	 * 		foo(  foo  )
	 * otherwise returns
	 * 		foo
	 * 
	 * @param str
	 * @return
	 */
	protected static String cleanStringForReport(String str) {
		if (sStringCleaningForReportFlag == false) {
			return str;
		}
		Object[] result = cleanStringBase(str);
		if (Boolean.FALSE.equals(result[0])) {
			return (String)result[1]; //the original value
		}
		return (String)(result[2] + "(" + result[1] + ")"); // foo(   foo   )
	}

    // ------------------------------------------------------
    // ----- errors
    // ------------------------------------------------------
    
	
    /**
     * Adds an error to the error list
     * 
     * @param error
     */
    public void addError(String error) {
    	if (errors == null) {
    		errors = new LinkedList<String>();
    	}
    	errors.add(error);
    }
	
    
    /**
     * Check if there are any errors
     * 
     * @return
     */
    public boolean doErrorsExist() {
    	if (errors==null || errors.size()==0) {
    		return false;
    	}
    	return true;
    }

    /**
     * Returns the list of errors
     * 
     * @return
     */
    public List<String> getErrors() {
        return errors;
    }

    
    
    
    /**
     * Returns the list of debug messages - CLASS & STATIC
     * 
     * @return
     */
    public List<String> getDebugs() {
        if (debugs == null || debugs.size()==0) {
            return null;
            
        }
        return debugs;
    }
    
    
    
    /**
     * Adds debug info to the debug list - CLASS INSTANCE
     * 
     * @param error
     */
    public void addDebug(String info) {
        if (debugs == null) {
            debugs = new LinkedList<String>();
        }
        debugs.add(info);
    }
    public void addDebug(String key, Object value) {
        addDebug(key + " = " + value);
    }

    
    
    /**
     * Check if there are any debug messages
     * 
     * @return
     */
    public boolean doDebugsExist() {
        if (debugs != null && 0 < debugs.size()) {
            return true;
        }
        return false;
    }

    /**
     * Formats an error message for surefire & DetailReport.
     * Includes debug info.
     * 
     * @return
     */
    protected String getErrorMsgIfExists() {
        String msg = null;
        if (doErrorsExist()) {          
                StringBuffer sb = new StringBuffer();

                //
                // check if the client only wants limited amount of info for the surefire report
                if (ErrorOption.SUREFIRE_SHORT.equals(sErrorOption)) {
                    sb.append("\n<br>" + errors.get(0));
                    sb.append("\n... more ... see target/QATools.org-detailReport.html");
                } else { //show all errors
                    formatErrors();
                    for (String str : errors) {
                        sb.append(str);
                    }
                }
    
                
                if (doDebugsExist()) {
                    if (DebugOption.EVERYWHERE.equals(sDebugOption)) {		// client may opt to limit debug info that shows in surefire
                        formatDebugs();
                        for (String str : debugs) {
                            sb.append(str);
                        }
                    }
                }
                sb.append("\n");
                msg = sb.toString();
        }  // if doErrorsExist
        return msg;
    }

    // ------------------------------------------------------
    // ----- 
    // ------------------------------------------------------
    
    /**
     * 
     */
    public void junitAssertNoErrors() throws Exception {
    	throwExceptionIfErrors();
    }
    
    /**
     * 
     */
    public void testNGAssertNoErrors() throws Exception {
    	throwExceptionIfErrors();
    }

    
    /**
     * If any errors where captured, throw an Exception
     */
    public void throwExceptionIfErrors() throws Exception {
        String msg = getErrorMsgIfExists();
        if (msg != null) {
            throw new Exception(msg);
        }
    }


    /**
     * If any errors where captured, throw an Exception
     */
    public void throwRuntimeExceptionIfErrors() {
        String msg = getErrorMsgIfExists();
		if (msg != null) {
			throw new RuntimeException(msg);
		}
    }


}